/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

/**
 * @file doca_mmap.h
 * @page doca mmap
 * @defgroup MMAP mmap define
 *
 * @{
 */

#ifndef DOCA_MMAP_H_
#define DOCA_MMAP_H_

#include <stddef.h>
#include <stdint.h>

#include <doca_dev.h>

#ifdef __cplusplus
extern "C" {
#endif

/*******************************************************************************
 * DOCA Memory Map
 ******************************************************************************/

/**
 * @brief Collection of memory ranges for both local & remote systems.
 *
 * Definitions:
 * Memory range - virtually contiguous fracture of memory space defined by address and length.
 * Chunk - local system memory range.
 *
 * Limitations:
 * - Memory map should not contain memory ranges from different systems.
 *
 *
 * +---------------------+          +-----------------------------------------+
 * |Memory map           |--------->| chunk 1       +------+  +----------+    |
 * |                     |\         |               | data |  | data     |    |
 * |                     | \        |               +------+  +----------+    |
 * +---------+           |  \       +-----------------------------------------+
 * | dev1    |           |   \
 * +---------+           |    \     +------------------------------------+
 * | dev2    |           |     ---->| chunk 2  +------+    +----------+  |
 * +---------+           |          |          | data |    |  data    |  |
 * |                     |          |          +------+    +----------+  |
 * +---------------------+          +------------------------------------+
 */
struct doca_mmap;

enum doca_mmap_access_flags {
	DOCA_MMAP_ACCESS_LOCAL_WRITE		= 1,
	DOCA_MMAP_ACCESS_REMOTE_WRITE		= (1<<1),
	DOCA_MMAP_ACCESS_REMOTE_READ		= (1<<2),
	DOCA_MMAP_ACCESS_REMOTE_ATOMIC		= (1<<3),
};

enum doca_mmap_property {
	DOCA_MMAP_UNKNOWN = 0,
	DOCA_MMAP_NAME,			/**< Memory map name set on creation. Read only property.
					  *  Type: char[32] */
	DOCA_MMAP_MAX_NUM_CHUNKS,	/**< Max number of chunks to populate.
					  *  The defult value is 1 if not set by the user before doca_mmap_start.
					  *  Once this limit reached further chunks populations will fail.
					  *  Type: uint32_t */
	DOCA_MMAP_MAX_NUM_DEVICES,	/**< Max number of devices to add to the mmap.
					  *  The defult value is 1 if not set by the user before doca_mmap_start.
					  *  Once this limit reached further devices additions will fail.
					  *  Type: uint32_t */
	DOCA_MMAP_NUM_BUFS,		/**< Total number of `struct doca_buf` objects pointing to the memory in the
					  *  mmap. Read only property.
					  *  Type: uint32_t */
	DOCA_MMAP_EXPORTED,		/**< True for exported memory map, false otherwise. Read only property.
					  *  Type: bool */
	DOCA_MMAP_FROM_EXPORT,		/**< True for memory map that was created from export, false otherwise.
					  *  Read only property.
					  *  Type: bool */
	DOCA_MMAP_ACCESS,		/**< Memory map access flags. use enum doca_mmap_access_flags.
					  *  The defult value is -
					  *  (DOCA_MMAP_ACCESS_LOCAL_WRITE |
					  *   DOCA_MMAP_ACCESS_REMOTE_WRITE | DOCA_MMAP_ACCESS_REMOTE_READ)
					  *  if not set by the user before doca_mmap_start.
					  *  Type: uint32_t */
	DOCA_MMAP_LAST,
};


/**
 * @brief Allocates zero size memory map object with default/unset attributes.
 *
 * @details The returned memory map object can be manipulated with
 * doca_mmap_property_set() API.
 *
 * Once all required mmap attributes set it should be reconfigured
 * and adjusted to meet object size setting with doca_mmap_start()
 * See doca_mmap_start for the rest of the details.
 *
 * @param [in] name
 * Name of newly created doca_mmap. The name of the mmap must not exceed 31 characters.
 * @param [out] mmap
 * DOCA memory map structure with default/unset attributes.
 *
 * @return
 * DOCA_SUCCESS - in case of success.
 * doca_error code - in case of failure:
 * - DOCA_ERROR_INVALID_VALUE - if an invalid input had been received.
 * - DOCA_ERROR_NO_MEMORY - failed to alloc doca_mmap.
 * - DOCA_ERROR_UNKNOWN - failed to set doca_mmap name.
 */
__DOCA_EXPERIMENTAL
doca_error_t doca_mmap_create(const char *name, struct doca_mmap **mmap);

/**
 * @brief Destroy DOCA Memory Map structure.
 *
 * @details Before calling this function all allocated buffers should be returned back to the mmap.
 *
 * @param [in] mmap
 * The DOCA memory map structure.
 *
 * @return
 * DOCA_SUCCESS - in case of success.
 * doca_error code - in case of failure:
 * - DOCA_ERROR_INVALID_VALUE - if an invalid input had been received.
 * - DOCA_ERROR_NOT_PERMITTED - if there is a memory region pointed by one or more `struct doca_buf`,
 *                              or if memory deregistration failed.
 */
__DOCA_EXPERIMENTAL
doca_error_t doca_mmap_destroy(struct doca_mmap *mmap);

/**
 * @brief Set the value of a DOCA Memory Map property.
 * Note: once a memory map object has been first started this functionality will not be available.
 *
 * @param [in] mmap
 * The DOCA memory map structure.
 * @param [in] property
 * The requested property to set. See enum doca_mmap_property.
 * @param [in] value
 * The new value of the property.
 * @param [in] size
 * The size of the property in bytes.
 *
 * @return
 * DOCA_SUCCESS - in case of success.
 * doca_error code - in case of failure:
 * - DOCA_ERROR_INVALID_VALUE - if an invalid input had been received.
 * - DOCA_ERROR_NOT_PERMITTED - if trying to set properties after first start of the mmap.
 */
__DOCA_EXPERIMENTAL
doca_error_t doca_mmap_property_set(struct doca_mmap *mmap, enum doca_mmap_property property,
				    const uint8_t *value, uint32_t size);
/**
 * @brief Get the value of a DOCA Memory Map property.
 *
 * @param [in] mmap
 * The DOCA memory map structure.
 * @param [in] property
 * The requested property to set. See enum doca_mmap_property.
 * @param [out] value
 * The current value of the property.
 * @param [in] size
 * The size of the property in bytes.
 *
 * @return
 * DOCA_SUCCESS - in case of success.
 * doca_error code - in case of failure:
 * - DOCA_ERROR_INVALID_VALUE - if an invalid input had been received.
 */
__DOCA_EXPERIMENTAL
doca_error_t doca_mmap_property_get(struct doca_mmap *mmap, enum doca_mmap_property property,
				    uint8_t *value, uint32_t size);

/**
 * @brief Start DOCA Memory Map.
 *
 * @details Un-started/stopped memory map object rejects all attempts to map doca_buf structures to its memory ranges.
 * Once memory map object started it can provide following functionality:
 * - doca_mmap_dev_add() & doca_mmap_dev_rm()
 * - doca_mmap_populate()
 * On first start verifies & finalizes the mmap object configuration, the user can no longer set mmap properties.
 *
 * @param [in] mmap
 * DOCA memory map structure.
 *
 * @return
 * DOCA_SUCCESS - in case of success.
 * doca_error code - in case of failure:
 * - DOCA_ERROR_INVALID_VALUE - if an invalid input had been received.
 * - DOCA_ERROR_NO_MEMORY - if memory allocation failed.
 */
__DOCA_EXPERIMENTAL
doca_error_t doca_mmap_start(struct doca_mmap *mmap);

/**
 * @brief Stop DOCA Memory Map.
 *
 * @details Un-started/stopped memory map object rejects all attempts to map doca_buf structures to its memory ranges.
 * For details see doca_mmap_start().
 *
 * @param [in] mmap
 * DOCA memory map structure.
 *
 * @return
 * DOCA_SUCCESS - in case of success.
 * doca_error code - in case of failure:
 * - DOCA_ERROR_INVALID_VALUE - if an invalid input had been received.
 */
__DOCA_EXPERIMENTAL
doca_error_t doca_mmap_stop(struct doca_mmap *mmap);

/**
 * @brief Register DOCA memory map on a given device.
 *
 * @details Should not be called on:
 * - un-started/stopped memory map object.
 * - exported/from_export memory map object.
 *
 * @param [in] mmap
 * DOCA memory map structure.
 * @param [in] dev
 * DOCA Dev instance with appropriate capability.
 *
 * @return
 * DOCA_SUCCESS - in case of success.
 * doca_error code - in case of failure:
 * - DOCA_ERROR_INVALID_VALUE - if an invalid input had been received.
 * - DOCA_ERROR_NOT_PERMITTED - if doca_mmap status is invalid for this operation or memory registration failed.
 * - DOCA_ERROR_NO_MEMORY - if reached to DOCA_MMAP_MAX_NUM_DEVICES.
 * - DOCA_ERROR_IN_USE - if doca_dev already exists in doca_mmap.
 */
__DOCA_EXPERIMENTAL
doca_error_t doca_mmap_dev_add(struct doca_mmap *mmap, struct doca_dev *dev);

/**
 * @brief Deregister given device from DOCA memory map.
 *
 * @details Should not be called on:
 * - un-started/stopped memory map object.
 * - exported/from_export memory map object.
 *
 * @param [in] mmap
 * DOCA memory map structure.
 * @param [in] dev
 * DOCA Dev instance that was previously added.
 *
 * @return
 * DOCA_SUCCESS - in case of success.
 * doca_error code - in case of failure:
 * - DOCA_ERROR_INVALID_VALUE - if an invalid input had been received or doca_dev doesn't exists in doca_mmap.
 * - DOCA_ERROR_NOT_PERMITTED - if doca_mmap status is invalid for this operation or memory deregistration failed.
 */
__DOCA_EXPERIMENTAL
doca_error_t doca_mmap_dev_rm(struct doca_mmap *mmap, struct doca_dev *dev);

/**
 * @brief Function to be called for each populated memory range on memory map destroy.
 *
 * @param[in] addr
 *   Memory range pointer.
 * @param[in] len
 *   Memory range length.
 * @param[in] opaque
 *   An opaque pointer passed to iterator.
 */
typedef void (doca_mmap_memrange_free_cb_t)(char *addr, size_t len, void *opaque);

/**
 * @brief Add memory range to DOCA memory map.
 *
 * @details Should not be called on:
 * - un-started/stopped memory map object.
 * - exported/from_export memory map object.
 *
 * @param [in] mmap
 * DOCA memory map structure.
 * @param [in] addr
 * Start address of the memory range to be populated.
 * @param [in] len
 * The size of the memory range in bytes.
 * @param [in] pg_sz
 * Page size alignment of the provided memory range.
 * @param [in] free_cb
 * Callback function to free the populated memory range on memory map destroy.
 * @param [in] opaque
 * Opaque value to be passed to free_cb once called.
 *
 * @return
 * DOCA_SUCCESS - in case of success.
 * doca_error code - in case of failure:
 * - DOCA_ERROR_INVALID_VALUE - if an invalid input had been received.
 * - DOCA_ERROR_NOT_PERMITTED - if doca_mmap status is invalid for this operation or device registration failed
 *                              or addr and len intersect with an existing chunk.
 * - DOCA_ERROR_NO_MEMORY - if reached to DOCA_MMAP_MAX_NUM_CHUNKS, or memory allocation failed.
 */
__DOCA_EXPERIMENTAL
doca_error_t doca_mmap_populate(struct doca_mmap *mmap, char *addr, size_t len, size_t pg_sz,
				doca_mmap_memrange_free_cb_t *free_cb, void *opaque);

/**
 * @brief Compose memory map representation for later import with
 * doca_mmap_create_from_export() for one of the devices previously added to
 * the memory map.
 *
 * @details Once this function called on the object it considered as exported.
 *
 * Freeing memory buffer pointed by `*export_desc` is the caller responsibility.
 *
 * Should not be called on:
 * - un-started/stopped memory map object.
 * - exported/from_export memory map object.
 *
 * @param [in] mmap
 * DOCA memory map structure.
 * @param [in] dev
 * Device previously added to the memory map via doca_mmap_dev_add().
 * @param [out] export_desc
 * On successful return should have a pointer to the allocated blob containing serialized representation of the memory
 * map object for the device provided as `dev`.
 * @param [out] export_desc_len
 * Length in bytes of the export_desc.
 *
 * @return
 * DOCA_SUCCESS - in case of success.
 * doca_error code - in case of failure:
 * - DOCA_ERROR_INVALID_VALUE - if an invalid input had been received or device does not exists in mmap.
 * - DOCA_ERROR_NOT_PERMITTED - if doca_mmap status is invalid for this operation.
 * The following errors will occur if failed to produce export descriptor:
 * - DOCA_ERROR_NO_MEMORY - if failed to alloc memory for export_desc.
 * - DOCA_ERROR_NOT_SUPPORTED
 * - DOCA_ERROR_UNKNOWN
 *
 * Limitation: Can only support mmap consisting of a single chunk.
 */
__DOCA_EXPERIMENTAL
doca_error_t doca_mmap_export(struct doca_mmap *mmap, const struct doca_dev *dev, uint8_t **export_desc,
			      size_t *export_desc_len);

/**
 * @brief Creates a memory map object representing memory ranges in remote system
 * memory space.
 *
 * @details Once this function called on the object it considered as from_export.
 *
 * @param [in] name
 * Name of newly created DOCA memory map.
 * @param [in] export_desc
 * An export descriptor generated by doca_mmap_export.
 * @param [in] export_desc_len
 * Length in bytes of the export_desc.
 * @param [in] dev
 * A local device connected to the device that resides in the exported mmap.
 * @param [out] mmap
 * DOCA memory map granting access to remote memory.
 *
 * @note: The created object not backed by local memory.
 *
 * @return
 * DOCA_SUCCESS - in case of success.
 * doca_error code - in case of failure:
 * - DOCA_ERROR_INVALID_VALUE - if an invalid input had been received or internal error.
 * The following errors are internal and will occur if failed to produce new mmap from export descriptor:
 * - DOCA_ERROR_NO_MEMORY - if internal memory allocation failed.
 * - DOCA_ERROR_NOT_SUPPORTED
 * - DOCA_ERROR_NOT_PERMITTED
 * - DOCA_ERROR_UNKNOWN
 *
 * Limitation: Can only support mmap consisting of a single chunk.
 */
__DOCA_EXPERIMENTAL
doca_error_t doca_mmap_create_from_export(const char *name, const uint8_t *export_desc, size_t export_desc_len,
					  struct doca_dev *dev, struct doca_mmap **mmap);


#ifdef __cplusplus
}
#endif

/** @} */

#endif /* DOCA_MMAP_H_ */
