/*
 * Copyright (c) 2021-2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

/**
 * @file doca_regex.h
 * @page doca regex
 * @defgroup RegEx RegEx engine
 * DOCA RegEx library. For more details please refer to the user guide on DOCA
 * devzone.
 *
 * @{
 */

#ifndef DOCA_REGEX_H_
#define DOCA_REGEX_H_

#ifdef __cplusplus
extern "C" {
#endif

#include <stdbool.h>
#include <stddef.h>
#include <stdint.h>

#include <doca_compat.h>
#include <doca_regex_buffer.h>
#include <doca_regex_device.h>
#include <doca_regex_mempool.h>

/**
 * Opaque structure representing a DOCA RegEx instance.
 */
struct doca_regex;

/**
 * Data required to dispatch a job to a RegEx engine.
 */
struct doca_regex_job_request {
	/** ID of the job, useful to correlate the response. */
	uint64_t id;

	/**
	 * IDs which can be used to select which group of rules are used to process this job. Set each value to a non
	 * zero value to enable this feature or 0 to ignore it.
	 */
	uint16_t rule_group_ids[4];

	/** Data for the job */
	struct doca_regex_buffer const *buffer;
};

/**
 * Description of a RegEx match
 */
struct doca_regex_match {
	/** Allows matches to be linked together for easy management and iteration */
	struct doca_regex_match *next;

	/** Index relative to the start of the job / stream where the match begins */
	uint32_t match_start;

	/** ID of rule used to generate this match. */
	uint32_t rule_id;

	/** Length of matched value. */
	uint32_t length;
};

/**
 * Response status flags
 */
enum doca_regex_status_flag {
	DOCA_REGEX_STATUS_SEARCH_FAILED = 1,
};

/**
 * Result of a RegEx search
 */
struct doca_regex_job_response {
	/** ID of the enqueued job. */
	uint64_t id;

	/** Response flags. A bit masked field for zero or more status flags. See doca_regex_status_flag */
	uint64_t status_flags;

	/** Total number of detected matches. */
	uint32_t detected_matches;

	/** Total number of returned matches. */
	uint32_t num_matches;

	/** Returned matches. Contains num_matches elements as a linked list */
	struct doca_regex_match *matches;
};

/**
 * Create a DOCA RegEx instance.
 *
 * @return
 * Non NULL doca_regex object on success, NULL otherwise.
 */
__DOCA_EXPERIMENTAL
struct doca_regex *doca_regex_create(void);

/**
 * Destroy DOCA RegEx instance.
 *
 * @param [in] regex
 * Instance to be destroyed, MUST NOT BE NULL.
 *
 * @note
 * The application must call doca_regex_stop() before destroying a running instance.
 */
__DOCA_EXPERIMENTAL
void doca_regex_destroy(struct doca_regex *regex);

/**
 * Make the RegEx Engine ready for processing jobs.
 *
 * @details After calling this functions the engine will not accept any further configuration but will start accepting
 * jobs. If further configuration is required the engine should be stopped first to allow this configuration to be
 * performed.
 *
 * @param [in] regex
 * The DOCA RegEx instance to start, MUST NOT be NULL..
 *
 * @return
 * 0 on success or a negative posix status code.
 */
__DOCA_EXPERIMENTAL
int doca_regex_start(struct doca_regex *regex);

/**
 * Stop the RegEx engine.
 *
 * @details Disable processing of jobs and return the engine to an idle state. The engine will then be ready to accept
 * new configuration or be ready to be destroyed. Any in flight jobs when this function is called will be terminated and
 * will not return any results.
 *
 * @param [in] regex
 * The DOCA RegEx instance to stop, MUST NOT be NULL.
 *
 * @return
 * 0 on success or a negative posix status code.
 */
__DOCA_EXPERIMENTAL
int doca_regex_stop(struct doca_regex *regex);

/**
 * Set the device to be used for hardware accelerated searches
 *
 * @param [in] regex
 * The RegEx engine instance to use.
 *
 * @param [in] device
 * Provide a non NULL value to register a device to use for hardware accelerated searches. The application retains
 * ownership of the device object. The Application shall ensure the lifetime of the device is maintained until the
 * device is either de-registered or the doca_regex instance is destroyed. Pass a NULL value to this function to
 * de-register a registered device.
 *
 * @return
 * 0 on success or a negative status code on failure.
 *
 * @note this function shall only be called when the doca_regex instance is not running. see doca_regex_start,
 * doca_regex_stop
 *
 * @note this function is not thread safe. Application must ensure that this is only called from a single thread.
 */
__DOCA_EXPERIMENTAL
int doca_regex_hw_device_set(struct doca_regex *regex, struct doca_regex_device *device);

/**
 * Set the device to be used for software based searches
 *
 * @param [in] regex
 * The RegEx engine instance to use.
 *
 * @param [in] device
 * Provide a non NULL value to register a device to use for software based searches. The application retains ownership
 * of the device object. The Application shall ensure the lifetime of the device is maintained until the device is
 * either de-registered or the doca_regex instance is destroyed. Pass a NULL value to this function to de-register a
 * registered device.
 *
 * @return
 * 0 on success or a negative status code on failure.
 *
 * @note this function shall only be called when the doca_regex instance is not running. see doca_regex_start,
 * doca_regex_stop
 *
 * @note this function is not thread safe. Application must ensure that this is only called from a single thread.
 */
__DOCA_EXPERIMENTAL
int doca_regex_sw_device_set(struct doca_regex *regex, struct doca_regex_device *device);

/**
 * Set the size of overlap to use when a job exceeds a devices maximum search size.
 *
 * @details
 * When a submitted job is larger than the receiving device can support it must be fragmented. This can cause issues if
 * a match exists but is split across two fragments. To remedy this an overlap size can be set so that these matches may
 * be detected. The overlap defined by this function specifies how many bytes of the previous search fragment will be
 * resent as part of the next search fragment. So for example if a 100 byte job is submitted and a device supported a 32
 * byte maximum job length then the jobs sent would look as follows:
 *
 * Overlap size   First job   Second Job   Third Job   Fourth job   Fifth Job   Sixth Job
 * 0              [0-31]      [32-63]      [64-95]     [96-99]      ---         ---
 * 8              [0-31]      [24-55]      [42-79]     [72-99]      ---         ---
 * 16             [0-31]      [16-47]      [32-63]     [48-79]      [64-95]     [80-99]
 *
 * This allows the user to select an overlap value which provides enough overlap to detect any match they must find for
 * the lowest cost.
 *
 * @note
 * The maximum overlap size is the smallest of the registered devices maximum job length minus one. This is validated
 * during doca_regex_start.
 *
 * @param [in] regex
 * The RegEx engine instance to use.
 *
 * @param [in] nb_overlap_bytes
 * Number of bytes of overlap to use.
 *
 * @return
 * 0 on success or a negative status code on failure.
 *
 * @note this function shall only be called when the doca_regex instance is not running. see doca_regex_start,
 * doca_regex_stop
 *
 * @note this function is not thread safe. Application must ensure that this is only called from a single thread.
 */
__DOCA_EXPERIMENTAL
int doca_regex_overlap_size_set(struct doca_regex *regex, uint16_t nb_overlap_bytes);

/**
 * Define a threshold for "small jobs" where it is preferable to scan them using a software RegEx device instead of the
 * hardware RegEx device. This can be useful in cases where a particular hardware device incurs more overhead in the
 * process of sending the job to and from hardware. The value that should be set here will vary depending on the
 * capability of the hardware device in use.
 *
 * @note Defaults to 256.
 *
 * @note This value is only considered when both software and hardware RegEx devices are configured.
 *
 * @param [in] regex
 * The RegEx engine instance to use.
 *
 * @param [in] threshold
 * Number of bytes, when a job has less bytes then threshold the engine will prefer to send the job to the software
 * RegEx device.
 *
 * @return
 * 0 on success or a negative status code on failure.
 *
 * @note this function shall only be called when the doca_regex instance is not running. see doca_regex_start,
 * doca_regex_stop
 *
 * @note this function is not thread safe. Application must ensure that this is only called from a single thread.
 */
__DOCA_EXPERIMENTAL
int doca_regex_small_job_sw_offload_threshold_set(struct doca_regex *regex, uint16_t threshold);

/**
 * Enable or disable software fallback. When a job is unable to be executed on a hardware RegEx device the engine can
 * automatically re-execute the job on the software device.
 *
 * @note Defaults to false (off).
 *
 * @note This value is only considered when both software and hardware RegEx devices are configured.
 *
 * @param [in] regex
 * The RegEx engine instance to use.
 *
 * @param [in] enabled
 * Feature is enabled: true = enabled, false = disabled
 *
 * @return
 * 0 on success or a negative status code on failure.
 *
 * @note this function shall only be called when the doca_regex instance is not running. see doca_regex_start,
 * doca_regex_stop
 *
 * @note this function is not thread safe. Application must ensure that this is only called from a single thread.
 */
__DOCA_EXPERIMENTAL
int doca_regex_sw_fallback_enabled_set(struct doca_regex *regex, bool enabled);

/**
 * Specify how many qp's are available for use
 *
 * @param [in] regex
 * The RegEx engine instance to use.
 *
 * @param [in] num_qps
 * Number of QP's that can be used. Must be at least 1.
 *
 * @return
 * 0 on success or a negative status code on failure.
 *
 * @note this function shall only be called when the doca_regex instance is not running. see doca_regex_start,
 * doca_regex_stop
 *
 * @note this function is not thread safe. Application must ensure that this is only called from a single thread.
 */
__DOCA_EXPERIMENTAL
int doca_regex_num_qps_set(struct doca_regex *regex, uint16_t num_qps);

/**
 * Register a memory pool to the engine so it can acquire doca_regex_match objects without requiring memory allocations.
 * If an application does not desire to get match details from a search then this can be left out. The application
 * retains ownership of the memory pool and must ensure the memory pool remains available until either the doca regex
 * engine is destroyed or it can be de-registered while the engine is stopped.
 *
 * @note This mempool will be used to allocate all matches for all jobs on the registered queue so it must be sized
 * sufficiently large.
 *
 * @note The application must return the matches in each doca_regex_job_response to the mempool once they no longer
 * needed. See doca_regex_mempool_obj_put.
 *
 * @param [in] regex
 * The RegEx engine instance.
 *
 * @param [in] mp
 * Provide a non NULL value to register a memory pool. Provide NULL to de-register it.
 *
 * @param [in] qid
 * ID of the queue to associate this mempool with
 *
 * @return
 * 0 on success or a negative status code on failure.
 *
 * @note this function shall only be called when the doca_regex instance is not running. see doca_regex_start,
 * doca_regex_stop
 *
 * @note this function is not thread safe. Application must ensure that this is only called from a single thread.
 */
__DOCA_EXPERIMENTAL
int doca_regex_qp_mempool_set(struct doca_regex *regex, struct doca_regex_mempool *mp, uint16_t qid);

/**
 * Program the given rules buffer(s) into the DOCA RegEx engine.
 *
 * @param [in] regex
 * The RegEx instance to program.
 *
 * @param [in] hw_compiled_rules_bin
 * Buffer of pre-compiled binary rules data suitable for use by the selected hw impl.
 *
 * @param [in] hw_compiled_rules_size
 * Number of bytes of data in the buffer pointed to by hw_compiled_rules_bin.
 *
 * @param [in] sw_compiled_rules_bin
 * Buffer of pre-compiled binary rules data suitable for use by the selected sw impl.
 *
 * @param [in] sw_compiled_rules_size
 * Number of bytes of data in the buffer pointed to by sw_compiled_rules_bin.
 *
 * @return
 * 0 on success of writing each rule to at least one regex_impl, or a negative posix status code.
 */
__DOCA_EXPERIMENTAL
int doca_regex_program_compiled_rules(struct doca_regex *regex, char const *hw_compiled_rules_bin,
				      size_t hw_compiled_rules_size, char const *sw_compiled_rules_bin,
				      size_t sw_compiled_rules_size);

/**
 * Program the given rules file(s) into the DOCA RegEx engine.
 *
 * @param [in] regex
 * The RegEx instance to program.
 *
 * @param [in] hw_compiled_rules_path
 * Path to a pre-compiled binary rules data suitable for use by the selected hw impl.
 *
 * @param [in] sw_compiled_rules_path
 * Path to a pre-compiled binary rules data suitable for use by the selected sw impl.
 *
 * @return
 * 0 on success of writing each rule to at least one regex_impl, or a negative posix status code.
 */
__DOCA_EXPERIMENTAL
int doca_regex_program_compiled_rules_file(struct doca_regex *regex, char const *hw_compiled_rules_path,
					   char const *sw_compiled_rules_path);

/**
 * Enqueue a job.
 *
 * @note
 * This function is thread safe when each queue is written to by only one thread. If there is a chance that more than
 * one thread could write to the same queue then external synchronisation must be provided by the application to prevent
 * that circumstance.
 *
 * @param [in] regex
 * The RegEx engine instance to use.
 *
 * @param [in] qid
 * ID of queue to enqueue the job into.
 *
 * @param [in] job
 * The job to enqueue. Caller retains ownership of the job
 *
 * @param [in] allow_aggregation
 * When set the RegEx engine may choose to not begin processing this job immediately in an attempt to maximise overall
 * efficiency and throughput. When not set the RegEx engine must begin processing immediately, potentially reducing
 * latency. This allows an application to favour either throughput or latency. If in doubt it is recommended to favour
 * throughput.
 *
 * @return
 * [0..1] Number of jobs enqueued or a negative posix status code.
 */
__DOCA_EXPERIMENTAL
int doca_regex_enqueue(struct doca_regex *regex, uint16_t qid, struct doca_regex_job_request const *job,
		       bool allow_aggregation);

/**
 * Dequeue responses from the RegEx engine.
 *
 * @note
 * This function is thread safe when each queue is read by only one thread. If there is a chance that more than one
 * thread could read from to the same queue then external synchronisation must be provided by the application to prevent
 * that circumstance.
 *
 * @param [in] regex
 * The RegEx engine to dequeue data from.
 *
 * @param [in] qid
 * ID of queue to read results from.
 *
 * @param [in,out] responses
 * A pointer to an array of responses. The RegEx engine will place each dequeued result into this array until either the
 * maximum number of responses are returned or all available responses have been returned. Caller assumes ownership of
 * each response returned this way. The engine takes no care to clear any un-populated element of this array, it is the
 * caller responsibility to ensure they only read as many responses as was indicated in the functions return value.
 *
 * @param [in] max_results
 * Maximum number of results to return. the responses array MUST have capacity for at least this many elements.
 *
 * @return
 * [0..max_results] The number of responses dequeued or a negative posix status code.
 */
__DOCA_EXPERIMENTAL
int doca_regex_dequeue(struct doca_regex *regex, uint16_t qid, struct doca_regex_job_response *responses,
		       uint8_t max_results);

#ifdef __cplusplus
} /* extern "C" */
#endif

/** @} */

#endif /* DOCA_REGEX_H_ */
