/*
 * Copyright (c) 2021-2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

/**
 * DOCA RegEx buffer supporting different payload types.
 *
 * @details
 * All methods of this API provide assurances similar to the strong exception guarantee in that:<br>
 * - A function will complete in its entirety or not at all. They will NEVER leave an argument or the operated upon
 * object in an undetermined state.<br>
 * - In the case of failure the operated upon object and any arguments are not modified in any way and it is as if the
 * function call was never executed.<br>
 * - A failed function call may have written to a log source. This is not considered a change in any objects state.
 */

#ifndef DOCA_REGEX_BUFFER_H_
#define DOCA_REGEX_BUFFER_H_

#ifdef __cplusplus
extern "C" {
#endif

#include <stdint.h>

#include "doca_regex_device.h"

/**
 * A type to represent a buffer to be scanned by the regex engine.
 */
struct doca_regex_buffer {
	/**
	 * Pointer to job data, must remain valid until the response for this job is returned
	 */
	void const *address;

	/** Number of bytes of data in this job. */
	uint32_t length;

	/**
	 * Is there an mkey associated with this memory. 1 = yes, 0 = no. All other values are invalid.
	 */
	uint32_t has_mkey;

	/** Mkey value. Only valid when has_mkey is set. */
	uint32_t mkey;
};

/**
 * Allocate a memory region and get mkey to send data to HW device
 *
 * @param [in] regex_dev
 * The regex_device to associate memory region with
 *
 * @param [in] data
 * The address of the data to assign to memory region
 *
 * @param [in] data_len
 * The length of the data to assign to memory region
 *
 * @param [out] mkey
 * The mkey for HW device to access memory region
 *
 * @return
 * 0 on success or a negative posix status code.
 */
__DOCA_EXPERIMENTAL
int doca_regex_buffer_generate_mkey(struct doca_regex_device *regex_dev, char *data, long data_len, uint32_t *mkey);

/**
 * Free memory regions assigned to a device
 *
 * @param [in] regex_dev
 * The regex_device with mem regions to free
 */
__DOCA_EXPERIMENTAL
void doca_regex_buffer_release_mkey(struct doca_regex_device *regex_dev);

#ifdef __cplusplus
} /* extern "C" */
#endif

#endif /* DOCA_REGEX_BUFFER_H_ */
