/*
 * Copyright (c) 2021-2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

/**
 * Definition of an abstract implementation to be used by doca regex, Provides a separation between implementation of
 * regex routines for the higher level functionality of the doca regex api.
 */

#ifndef DOCA_REGEX_DEVICE_H_
#define DOCA_REGEX_DEVICE_H_

#ifdef __cplusplus
extern "C" {
#endif

#include <stdbool.h>
#include <stdint.h>

#include <doca_regex_mempool.h>

/**
 * Structure describing the capabilities supported by a regex impl.
 */
struct doca_regex_dev_caps {
	/** Device has PMI support, 0 if not supported. */
	uint8_t pmi_supported;

	/** Device has multi-client support, 0 if not supported. */
	uint8_t multi_client_supported;
};

/**
 * Structure describing a regex impl.
 */
struct doca_regex_dev_info {
	/** version of the regex device [MAJOR][MINOR][PATCH]. */
	uint16_t version[3];

	/** Maximum size of a job payload in bytes for a single job. */
	uint32_t max_job_size;

	/** A name describing this device. */
	char const *name;
};

/**
 * Structure describing the result of compilation of rules into a given regex devices preferred format.
 */
struct doca_regex_compiled_rules_result {
	/** Pointer to compiled binary rules data. */
	void *compiled_rules_data;

	/** Array of rule IDs of rules that failed to compile. length == num_failures. */
	uint32_t *failure_indices;

	/** Number of rules that successfully compiled. */
	uint32_t num_successful;

	/** Number of rules that failed to compile. */
	uint32_t num_failures;

	/** size of compiled binary rules data. */
	uint32_t compiled_rules_size;
};

/*
 * Forward type declarations
 */
struct doca_regex_job_request;
struct doca_regex_job_response;

/**
 * Structure containing all the functionality required of a regex device by the DOCA regex.
 *
 * @details
 * Allows each DOCA regex instance to be supported by any regex_impl. Such as Hyperscan, DPDK regex, or any other regex
 * library that can be adapted to this interface.
 */
struct doca_regex_device {
	/** A place to bind any required extra data into the impl. */
	void *cookie;

	/** Protection Domain assigned to device (only applicable to HW dev). */
	void *pd;

	/** Memory regions associated with the device (only applicable to HW dev). */
	void *mem_regs;

	/**
	 * @brief
	 * Do any one time setup work that may be required before performing regex operations.
	 *
	 * @param [in] inst
	 * Instance of the impl.
	 *
	 * @param [in] dev_addr
	 * PCIe address of regex device (NULL if not a HW device).
	 *
	 * @return
	 * 0 on success or a negative posix status code.
	 */
	int (*init_fn)(struct doca_regex_device *inst, const char *dev_addr);

	/**
	 * @brief
	 * Cleanup all resources acquired by the regex_impl from either one time setup or during runtime operation.
	 *
	 * @param [in] inst
	 * Instance of the impl.
	 *
	 * @return
	 * 0 on success or a negative posix status code.
	 */
	int (*cleanup_fn)(struct doca_regex_device *inst);

	/**
	 * @brief
	 * Function suitable to destroy this doca_regex_device.
	 *
	 * @param [in] inst
	 * Instance of the impl.
	 *
	 * @details
	 * Set this to NULL to indicate that the engine should not attempt to destroy this object when the engine is
	 * destroyed.
	 */
	void (*destroy_fn)(struct doca_regex_device *inst);

	/**
	 * @brief
	 * Query capabilities of the regex device.
	 *
	 * @param [in] inst
	 * Instance of the impl.
	 *
	 * @param [out] caps
	 * Capabilities structure to populate. MUST NOT BE NULL.
	 *
	 * @return
	 * 0 on success or a negative posix status code.
	 */
	int (*query_dev_caps_fn)(struct doca_regex_device *inst, struct doca_regex_dev_caps *caps);

	/**
	 * @brief
	 * Query information about the regex device.
	 *
	 * @param [in] inst
	 * Instance of the impl.
	 *
	 * @param [out] info
	 * Informational structure to populate. MUST NOT BE NULL.
	 *
	 * @return
	 * 0 on success or a negative posix status code.
	 */
	int (*query_dev_info_fn)(struct doca_regex_device *inst, struct doca_regex_dev_info *info);

	/**
	 * @brief
	 * Compile given raw rules.
	 *
	 * @param [in] inst
	 * Instance of the impl.
	 *
	 * @param [in] raw_rules
	 * A 2d array of strings which hold the raw rules in a text format.
	 *
	 * @param [in] num_lines
	 * Number of lines of text to parse.
	 *
	 * @return
	 * A doca_regex_compiled_rules_result describing the number of failed rules and outlining the index of each rule
	 * that failed. Caller assumes ownership of this object.
	 */
	struct doca_regex_compiled_rules_result *(*compile_rules_fn)(struct doca_regex_device *inst,
								     char const *const *raw_rules, uint32_t num_lines);

	/**
	 * @brief
	 * Program regex device with the supplied compiled rules data.
	 *
	 * @param [in] inst
	 * Instance of the impl.
	 *
	 * @param [in] data
	 * A blob of memory in the same format that would be produced by compile_rules_fn. However it may have been read
	 * from a pre-compiled input file and not necessarily compiled by this regex_impl. It will be the applications
	 * responsibility to ensure that when providing a path to a pre-compiled data within doca_regex_rules_cfg that
	 * the file contains data in a format suitable for the selected regex_impl.
	 *
	 * @param [in] data_len
	 * Number of bytes in the blob pointed to by data.
	 *
	 * @param [in] incremental
	 * Set to true when providing additional rules to a regex_impl, false to start from scratch.
	 *
	 * @return
	 * 0 on success or a negative posix status code.
	 */
	int (*program_compiled_rules_fn)(struct doca_regex_device *inst, void const *data, uint32_t data_len);

	/**
	 * @brief
	 * Configure the number of QP's required by the regex lib from the device.
	 *
	 * @param [in] inst
	 * Instance of the impl.
	 *
	 * @param [in] num_qps
	 * Required number of qps.
	 *
	 * @return
	 * 0 on success or a negative posix status code.
	 */
	int (*configure_nb_qps_fn)(struct doca_regex_device *inst, uint32_t num_qps);

	/**
	 * @brief
	 * Send a job to the regex engine to be processed.
	 *
	 * @param [in] inst
	 * Instance of the impl.
	 *
	 * @param [in] qpid
	 * ID of qp to use
	 *
	 * @param [in] job
	 * Details of the job. Caller retains ownership of these objects;
	 *
	 * @param [in] internal_job_id
	 * Internal job id - this value should be returned by the device when dequeuing the result for this piece. The
	 * job request id is the id of the overall job which may have been broken into multiple pieces.
	 *
	 * @param [in] allow_aggregation
	 * Give permission to the device to hold the job to potentially improve throughput.
	 * If false then the device must immediately send the job.
	 *
	 * @return
	 * [0..count] Number of jobs enqueued or a negative posix status code.
	 */
	int (*enqueue_job_fn)(struct doca_regex_device *inst, uint16_t qpid, struct doca_regex_job_request const *job,
			      uint64_t internal_job_id, bool allow_aggregation);

	/**
	 * @brief
	 * Fetch job responses from the regex_impl.
	 *
	 * @param [in] inst
	 * Instance of the impl.
	 *
	 * @param [in,out] responses
	 * A pointer to an array of responses. The regex engine will place each dequeued result into this array until
	 * either the maximum number of responses are returned or all available responses have been returned
	 *
	 * @param [in] max_responses
	 * The maximum number of responses the engine is able to accept from the regex_impl.
	 *
	 * @param [in] matches_mp
	 * To get matches back from the regex search provide a non null memory pool which will return items which are
	 * least as big as a doca_regex_match. Provide A NULL pool to just get a summary of the number of matches
	 * detected.
	 *
	 * @return
	 * [0..max_results] Number of jobs dequeued or a negative posix status code.
	 */
	int (*dequeue_jobs_fn)(struct doca_regex_device *inst, uint16_t qpid, struct doca_regex_job_response *responses,
			       uint8_t max_results, struct doca_regex_mempool *matches_mp);
};

/**
 * Get a new instance of a pre-configured implementation of a doca_regex_device.
 *
 * @details
 * DOCA regex provides some regex_impl implementations that are ready to use. This saves on the effort of developing
 * them yourself. You are free to disregard these and write your own to link doca_regex to another regex implementation
 * or if you need something custom to your needs.
 *
 * @param [in] name
 * Name of the pre-configured doca_regex_device to get an instance of.
 *
 * @return
 * A non NULL regex_impl on success, or NULL if the name is not recognised or some other error occurred.
 */
__DOCA_EXPERIMENTAL
struct doca_regex_device *doca_regex_create_pre_configured_regex_impl(char const *name);

/**
 * Convenience function to destroy a regex impl.
 *
 * @param [in] inst;
 * When this value is not NULL delegates the destruction to the impls destroy_fn
 *
 * @note
 * Will not call cleanup before destroying, it's assumed thats handled by the application.
 */
__DOCA_EXPERIMENTAL
void doca_regex_device_destroy(struct doca_regex_device *inst);

#ifdef __cplusplus
} /* extern "C" */
#endif

#endif /* DOCA_REGEX_DEVICE_H_ */
